"""The satellite is an open source library to work with the private assistant oecosystem built to run on edge
devices. It allows the other components to interact speech based with the user and listen for user keywords
to activate."""

try:
    from ._version import __version__
except ImportError:
    # Fallback for development installs
    __version__ = "dev"

__all__ = ["__version__"]



"""Basic tests for config module."""

import tempfile
from pathlib import Path

import pytest
from pydantic import ValidationError

from private_assistant_comms_satellite.utils.config import Config, load_config


class TestConfig:
    """Basic tests for Config model."""

    def test_config_defaults(self):
        """Test that Config initializes with expected defaults."""
        config = Config()

        assert config.wakework_detection_threshold == 0.6  # noqa: PLR2004
        assert config.path_or_name_wakeword_model == "./hey_nova.onnx"
        assert config.name_wakeword_model == "hey_nova"
        assert config.samplerate == 16000  # noqa: PLR2004
        assert config.chunk_size == 512  # noqa: PLR2004
        assert config.vad_threshold == 0.6  # noqa: PLR2004
        assert config.mqtt_server_host == "localhost"
        assert config.mqtt_server_port == 1883  # noqa: PLR2004

    def test_config_with_custom_values(self):
        """Test Config with custom values."""
        config = Config(
            wakework_detection_threshold=0.8,
            samplerate=22050,
            mqtt_server_host="custom.mqtt.server",
            mqtt_server_port=8883,
        )

        assert config.wakework_detection_threshold == 0.8  # noqa: PLR2004
        assert config.samplerate == 22050  # noqa: PLR2004
        assert config.mqtt_server_host == "custom.mqtt.server"
        assert config.mqtt_server_port == 8883  # noqa: PLR2004

    def test_config_topic_properties(self):
        """Test computed topic properties."""
        config = Config(client_id="test_client")

        expected_base = "assistant/comms_bridge/all/test_client"
        assert config.base_topic == expected_base
        assert config.input_topic == f"{expected_base}/input"
        assert config.output_topic == f"{expected_base}/output"

    def test_config_topic_overwrites(self):
        """Test topic overwrites work correctly."""
        config = Config(
            base_topic_overwrite="custom/base",
            input_topic_overwrite="custom/input",
            output_topic_overwrite="custom/output",
        )

        assert config.base_topic == "custom/base"
        assert config.input_topic == "custom/input"
        assert config.output_topic == "custom/output"


class TestLoadConfig:
    """Basic tests for config loading."""

    def test_load_valid_config(self):
        """Test loading a valid YAML config file."""
        config_data = """
wakework_detection_threshold: 0.7
samplerate: 22050
mqtt_server_host: "test.server"
mqtt_server_port: 8883
"""

        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(config_data)
            f.flush()

            config_path = Path(f.name)
            config = load_config(config_path)

            assert config.wakework_detection_threshold == 0.7  # noqa: PLR2004
            assert config.samplerate == 22050  # noqa: PLR2004
            assert config.mqtt_server_host == "test.server"
            assert config.mqtt_server_port == 8883  # noqa: PLR2004

            # Clean up
            config_path.unlink()

    def test_load_config_file_not_found(self):
        """Test loading non-existent config file raises FileNotFoundError."""
        non_existent_path = Path("does_not_exist.yaml")

        with pytest.raises(FileNotFoundError):
            load_config(non_existent_path)

    def test_load_invalid_config(self):
        """Test loading invalid config raises ValidationError."""
        invalid_config = """
wakework_detection_threshold: "not_a_number"
samplerate: -1000
"""

        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(invalid_config)
            f.flush()

            config_path = Path(f.name)

            with pytest.raises(ValidationError):
                load_config(config_path)

            # Clean up
            config_path.unlink()

    def test_load_empty_config_uses_defaults(self):
        """Test loading empty config file uses defaults."""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write("{}")  # Empty YAML
            f.flush()

            config_path = Path(f.name)
            config = load_config(config_path)

            # Should use defaults
            assert config.wakework_detection_threshold == 0.6  # noqa: PLR2004
            assert config.samplerate == 16000  # noqa: PLR2004

            # Clean up
            config_path.unlink()



import asyncio
import queue
import ssl

import aiomqtt
from private_assistant_commons import messages, skill_logger

logger = skill_logger.SkillLogger.get_logger("Private Assistant Comms Satellite")


class AsyncMQTTClient:
    """Async MQTT client wrapper for aiomqtt integration."""

    def __init__(  # noqa: PLR0913
        self,
        hostname: str,
        port: int,
        client_id: str,
        topic_to_queue: dict[str, asyncio.Queue[messages.Response]],
        use_websockets: bool = False,
        websocket_path: str = "/mqtt",
        use_ssl: bool = False,
    ):
        self.hostname = hostname
        self.port = port
        self.client_id = client_id
        self.topic_to_queue: dict[str, asyncio.Queue[messages.Response]] = topic_to_queue
        self.use_websockets = use_websockets
        self.websocket_path = websocket_path
        self.use_ssl = use_ssl
        self._client: aiomqtt.Client | None = None
        self._running = False

    async def _handle_client_connection(self, client: aiomqtt.Client) -> None:
        """Handle MQTT client connection, subscription and message processing."""
        self._client = client
        logger.info("Connected to MQTT server")

        # Subscribe to all topics
        for topic in self.topic_to_queue:
            await client.subscribe(topic, qos=1)
            logger.debug("Subscribed to topic: %s", topic)

        # Handle messages indefinitely
        async for message in client.messages:
            topic_queue = self.topic_to_queue.get(message.topic.value)
            logger.debug("Received message on topic %s: %s", message.topic, message.payload)

            if topic_queue is None:
                logger.warning("%s seems to have no queue. Discarding message.", message.topic.value)
            else:
                try:
                    # AIDEV-NOTE: Payload parsing handles various aiomqtt message formats for robustness
                    if isinstance(message.payload, bytes | bytearray):
                        payload_str = message.payload.decode("utf-8")
                    elif isinstance(message.payload, str):
                        payload_str = message.payload
                    else:
                        payload_str = str(message.payload)
                    topic_queue.put_nowait(messages.Response.model_validate_json(payload_str))
                except queue.Full:
                    logger.warning("Queue for topic %s is full, discarding message", message.topic.value)

    async def start(self) -> None:
        """Start the MQTT client and message handling."""
        if self._running:
            return

        self._running = True
        logger.info("Starting MQTT client connection to %s:%s", self.hostname, self.port)
        logger.info("WebSocket mode: %s, SSL mode: %s", self.use_websockets, self.use_ssl)
        if self.use_websockets:
            logger.info("WebSocket path: %s", self.websocket_path)

        try:
            # Configure SSL context if needed
            tls_context = None
            if self.use_ssl:
                tls_context = ssl.create_default_context()
                logger.info("Using SSL/TLS encryption")

            # Configure transport based on WebSocket setting
            if self.use_websockets:
                if self.use_ssl:
                    logger.info("Using secure WebSocket (WSS) transport with path: %s", self.websocket_path)
                else:
                    logger.info("Using WebSocket (WS) transport with path: %s", self.websocket_path)

                logger.debug("Creating WebSocket client with SSL context: %s", tls_context is not None)
                async with aiomqtt.Client(
                    hostname=self.hostname,
                    port=self.port,
                    identifier=self.client_id,
                    transport="websockets",
                    websocket_path=self.websocket_path,
                    tls_context=tls_context,
                    timeout=30,  # Add explicit timeout
                ) as client:
                    await self._handle_client_connection(client)
            else:
                if self.use_ssl:
                    logger.info("Using MQTT over SSL/TLS")

                async with aiomqtt.Client(
                    hostname=self.hostname,
                    port=self.port,
                    identifier=self.client_id,
                    tls_context=tls_context,
                ) as client:
                    await self._handle_client_connection(client)
        except Exception as e:
            logger.error("MQTT client error: %s", e)
        finally:
            self._running = False
            self._client = None

    async def publish(self, topic: str, payload: str, qos: int = 1) -> None:
        """Publish a message to the specified topic."""
        if not self._client:
            logger.error("MQTT client not connected, cannot publish")
            return

        try:
            await self._client.publish(topic, payload, qos=qos)
            logger.debug("Published message to %s: %s", topic, payload)
        except Exception as e:
            logger.error("Failed to publish message: %s", e)

    async def stop(self) -> None:
        """Stop the MQTT client."""
        self._running = False
        # AIDEV-NOTE: Client cleanup happens automatically in context manager
        logger.info("MQTT client stopped")



"""Basic tests for mqtt_utils module."""

import queue

import pytest

from private_assistant_comms_satellite.utils.mqtt_utils import AsyncMQTTClient


@pytest.fixture
def sample_topic_queue():
    """Create sample topic to queue mapping for testing."""
    return {
        "test/topic1": queue.Queue(),
        "test/topic2": queue.Queue(),
    }


class TestAsyncMQTTClientInit:
    """Test AsyncMQTTClient initialization."""

    def test_client_initialization(self, sample_topic_queue):
        """Test proper initialization of AsyncMQTTClient."""
        client = AsyncMQTTClient(
            hostname="localhost",
            port=1883,
            client_id="test_client",
            topic_to_queue=sample_topic_queue,
        )

        assert client.hostname == "localhost"
        assert client.port == 1883  # noqa: PLR2004
        assert client.client_id == "test_client"
        assert client.topic_to_queue == sample_topic_queue
        assert client.use_websockets is False
        assert client.websocket_path == "/mqtt"
        assert client._client is None
        assert client._running is False

    def test_client_initialization_with_websockets(self, sample_topic_queue):
        """Test proper initialization of AsyncMQTTClient with WebSocket transport."""
        client = AsyncMQTTClient(
            hostname="localhost",
            port=8080,
            client_id="test_client",
            topic_to_queue=sample_topic_queue,
            use_websockets=True,
            websocket_path="/custom/mqtt",
        )

        assert client.hostname == "localhost"
        assert client.port == 8080  # noqa: PLR2004
        assert client.client_id == "test_client"
        assert client.topic_to_queue == sample_topic_queue
        assert client.use_websockets is True
        assert client.websocket_path == "/custom/mqtt"
        assert client.use_ssl is False
        assert client._client is None
        assert client._running is False

    def test_client_initialization_with_ssl(self, sample_topic_queue):
        """Test proper initialization of AsyncMQTTClient with SSL/TLS transport."""
        client = AsyncMQTTClient(
            hostname="broker.example.com",
            port=8883,
            client_id="test_client",
            topic_to_queue=sample_topic_queue,
            use_ssl=True,
        )

        assert client.hostname == "broker.example.com"
        assert client.port == 8883  # noqa: PLR2004
        assert client.client_id == "test_client"
        assert client.topic_to_queue == sample_topic_queue
        assert client.use_websockets is False
        assert client.websocket_path == "/mqtt"
        assert client.use_ssl is True
        assert client._client is None
        assert client._running is False

    def test_client_initialization_with_websockets_and_ssl(self, sample_topic_queue):
        """Test proper initialization of AsyncMQTTClient with WebSocket Secure (WSS) transport."""
        client = AsyncMQTTClient(
            hostname="broker.example.com",
            port=443,
            client_id="test_client",
            topic_to_queue=sample_topic_queue,
            use_websockets=True,
            websocket_path="/secure/mqtt",
            use_ssl=True,
        )

        assert client.hostname == "broker.example.com"
        assert client.port == 443  # noqa: PLR2004
        assert client.client_id == "test_client"
        assert client.topic_to_queue == sample_topic_queue
        assert client.use_websockets is True
        assert client.websocket_path == "/secure/mqtt"
        assert client.use_ssl is True
        assert client._client is None
        assert client._running is False


# NOTE: Complex async integration tests removed for simplicity -
# focus on basic initialization and configuration testing



import asyncio
import os
import pathlib
import wave

import openwakeword
import openwakeword.utils
from private_assistant_commons import messages, skill_logger

from private_assistant_comms_satellite import satellite
from private_assistant_comms_satellite.utils import (
    config,
    mqtt_utils,
)


# Load WAV file into memory
def load_wav_file(file_path: str) -> bytes:
    """Load WAV file into memory using context manager."""
    with wave.open(file_path, "rb") as wf:
        return wf.readframes(wf.getnframes())


def ensure_openwakeword_models() -> None:
    """Ensure OpenWakeWord models are downloaded for first-time setup.

    Downloads spectrograms and silero VAD model if not already present.
    This prevents the need for manual model download before first use.
    """
    logger = skill_logger.SkillLogger.get_logger("Private Assistant Comms Satellite")

    logger.info("Ensuring OpenWakeWord base models are available...")
    try:
        # Download base models (spectrograms and silero VAD) - only downloads if missing
        openwakeword.utils.download_models(model_names=["none"])
        logger.debug("OpenWakeWord base models verified successfully")
    except Exception as e:
        logger.error("Failed to download OpenWakeWord models: %s", e)
        logger.error("Please download manually")
        raise


def start_satellite(config_path: pathlib.Path) -> None:
    """Start the satellite application with the given configuration.

    Args:
        config_path: Path to the YAML configuration file
    """
    # AIDEV-NOTE: Refactored from main() to allow CLI integration
    # Ensure base models are available before loading config and creating wakeword model
    ensure_openwakeword_models()

    config_obj = config.load_config(config_path)
    wakeword_model = openwakeword.Model(
        wakeword_models=[config_obj.path_or_name_wakeword_model],
        enable_speex_noise_suppression=True,
        vad_threshold=config_obj.vad_threshold,
        inference_framework=config_obj.openwakeword_inference_framework,
    )
    topic_to_queue: dict[str, asyncio.Queue[messages.Response]] = {}
    output_queue: asyncio.Queue[messages.Response] = asyncio.Queue()
    output_topic = config_obj.output_topic
    topic_to_queue[output_topic] = output_queue
    topic_to_queue[config_obj.broadcast_topic] = output_queue

    mqtt_client = mqtt_utils.AsyncMQTTClient(
        hostname=config_obj.mqtt_server_host,
        port=config_obj.mqtt_server_port,
        client_id=config_obj.client_id,
        topic_to_queue=topic_to_queue,
        use_websockets=config_obj.mqtt_use_websockets,
        websocket_path=config_obj.mqtt_websocket_path,
        use_ssl=config_obj.mqtt_use_ssl,
    )

    logger = skill_logger.SkillLogger.get_logger("Private Assistant Comms Satellite")
    # Preload sounds
    start_listening_sound = load_wav_file(config_obj.start_listening_path)
    stop_listening_sound = load_wav_file(config_obj.stop_listening_path)
    satellite_handler = satellite.Satellite(
        config=config_obj,
        output_queue=output_queue,
        start_listening_sound=start_listening_sound,
        stop_listening_sound=stop_listening_sound,
        wakeword_model=wakeword_model,
        mqtt_client=mqtt_client,
        logger=logger,
    )

    try:
        satellite_handler.start()
    except KeyboardInterrupt:
        logger.info("Interrupted by user, stopping...")
    finally:
        satellite_handler.cleanup()


if __name__ == "__main__":
    # AIDEV-NOTE: Direct execution for development, use CLI for production
    config_path = pathlib.Path(os.getenv("PRIVATE_ASSISTANT_API_CONFIG_PATH", "local_config.yaml"))
    start_satellite(config_path)



"""Voice activity detection."""

import numpy as np
from pysilero_vad import SileroVoiceActivityDetector


class SileroVad:
    """Voice activity detection with silero VAD."""

    def __init__(self, threshold: float, trigger_level: int) -> None:
        self.detector = SileroVoiceActivityDetector()
        self.threshold = threshold
        self.trigger_level = trigger_level
        self._activation = 0

        # AIDEV-NOTE: Pre-calculate chunk size and optimize for vectorized processing
        self._chunk_size = self.detector.chunk_bytes()
        self._chunk_array = np.zeros(self._chunk_size, dtype=np.uint8)  # Reusable chunk buffer

    def __call__(self, audio_bytes: bytes | None) -> bool:
        """Process audio chunk and detect voice activity with trigger mechanism.

        Args:
            audio_bytes: Audio data chunk or None to reset detector

        Returns:
            True if sustained speech detected above trigger level
        """
        # AIDEV-NOTE: VAD reset mechanism for initialization and error recovery
        if audio_bytes is None:
            # Reset
            self._activation = 0
            self.detector.reset()
            return False

        if len(audio_bytes) < self._chunk_size:
            raise ValueError(f"Audio bytes must be at least {self._chunk_size} bytes")

        # AIDEV-NOTE: Vectorized processing - process larger chunks efficiently
        return self._process_audio_vectorized(audio_bytes)

    def _process_audio_vectorized(self, audio_bytes: bytes) -> bool:
        """Vectorized audio processing for improved performance."""
        audio_len = len(audio_bytes)

        # AIDEV-NOTE: Calculate number of complete chunks to process
        num_chunks = audio_len // self._chunk_size

        if num_chunks == 0:
            return False

        # AIDEV-NOTE: Optimize for single chunk (most common case)
        if num_chunks == 1:
            prob = self.detector(audio_bytes[: self._chunk_size])
            return self._update_activation(prob)

        # AIDEV-NOTE: Vectorized processing for multiple chunks
        # Convert bytes to numpy array for efficient slicing
        audio_array = np.frombuffer(audio_bytes, dtype=np.uint8)

        # Process chunks in batches to reduce Python loop overhead
        batch_size = min(8, num_chunks)  # Process up to 8 chunks at once
        max_prob = 0.0

        for batch_start in range(0, num_chunks, batch_size):
            batch_end = min(batch_start + batch_size, num_chunks)
            batch_probs = []

            # Process batch of chunks
            for i in range(batch_start, batch_end):
                start_idx = i * self._chunk_size
                end_idx = start_idx + self._chunk_size

                # AIDEV-NOTE: Use array view instead of bytes slicing for efficiency
                chunk_view = audio_array[start_idx:end_idx]
                self._chunk_array[:] = chunk_view  # Copy to reusable buffer
                prob = self.detector(self._chunk_array.tobytes())
                batch_probs.append(prob)

            # AIDEV-NOTE: Use numpy max for efficient batch processing
            batch_max = np.max(batch_probs) if batch_probs else 0.0
            max_prob = max(max_prob, batch_max)

            # Early termination if we already found high probability
            if max_prob >= self.threshold:
                break

        return self._update_activation(max_prob)

    def _update_activation(self, probability: float) -> bool:
        """Update activation state based on probability."""
        # AIDEV-NOTE: Trigger-level mechanism reduces false positives in edge device environments
        if probability >= self.threshold:
            # Speech detected
            self._activation += 1
            if self._activation >= self.trigger_level:
                self._activation = 0
                return True
        else:
            # Silence detected
            self._activation = max(0, self._activation - 1)

        return False



from __future__ import annotations

import asyncio
import enum
import threading
import time
import uuid
from typing import TYPE_CHECKING, Any

import numpy as np
from private_assistant_commons import messages

# AIDEV-NOTE: Import pyaudio as optional dependency for CI/CD compatibility
if TYPE_CHECKING:
    import logging

    import openwakeword
    import pyaudio
else:
    try:
        import pyaudio
    except ImportError:
        pyaudio = None

from private_assistant_comms_satellite import silero_vad
from private_assistant_comms_satellite.utils import (
    config,
    mqtt_utils,
    speech_recognition_tools,
)


class SatelliteState(enum.Enum):
    """Satellite operation states."""

    LISTENING = "listening"  # Analyzing audio for wakeword
    RECORDING = "recording"  # Recording user command after wakeword
    WAITING = "waiting"  # Waiting for API response
    SPEAKING = "speaking"  # Playing TTS audio


class Satellite:
    """Main satellite class for voice interaction with simple state machine."""

    def __init__(  # noqa: PLR0913
        self,
        config: config.Config,
        output_queue: asyncio.Queue[messages.Response],
        start_listening_sound: bytes,
        stop_listening_sound: bytes,
        wakeword_model: openwakeword.Model,
        mqtt_client: mqtt_utils.AsyncMQTTClient,
        logger: logging.Logger,
    ):
        # Assign configuration
        self.config = config
        self.output_queue = output_queue
        # Assign preloaded WAV data
        self.start_listening_sound: bytes = start_listening_sound
        self.stop_listening_sound: bytes = stop_listening_sound
        self.vad_model = silero_vad.SileroVad(threshold=config.vad_threshold, trigger_level=config.vad_trigger)
        self.logger = logger
        # Assign wakeword model
        self.wakeword_model: openwakeword.Model = wakeword_model
        self.mqtt_client = mqtt_client
        self._mqtt_thread: threading.Thread | None = None
        self._mqtt_loop: asyncio.AbstractEventLoop | None = None

        # AIDEV-NOTE: Simple state machine
        self._state = SatelliteState.LISTENING
        self._state_lock = threading.RLock()

        # AIDEV-NOTE: Audio recording buffer
        max_recording_samples = self.config.max_command_input_seconds * self.config.samplerate
        self._audio_buffer = np.zeros(max_recording_samples, dtype=np.float32)
        self._buffer_position = 0

        # AIDEV-NOTE: Timing tracking
        self._stop_sound_time = 0.0

        # AIDEV-NOTE: PyAudio initialization - critical dependency for edge device audio I/O
        if pyaudio is None:
            raise ImportError(
                "PyAudio is required for audio processing: pip install 'private-assistant-comms-satellite[audio]'"
            )

        self.p: Any = pyaudio.PyAudio()
        self.stream_input: Any = self.p.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=self.config.samplerate,
            input=True,
            frames_per_buffer=self.config.chunk_size,
            input_device_index=self.config.input_device_index,
        )
        self.stream_output: Any = self.p.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=self.config.samplerate,
            output=True,
            output_device_index=self.config.output_device_index,
        )

    def _get_state(self) -> SatelliteState:
        """Thread-safe state getter."""
        with self._state_lock:
            return self._state

    def _set_state(self, new_state: SatelliteState) -> None:
        """Thread-safe state setter."""
        with self._state_lock:
            if self._state != new_state:
                self.logger.info("State change: %s -> %s", self._state.value, new_state.value)
                self._state = new_state

    async def _process_output_queue(self) -> None:
        """Process pending TTS responses from the assistant."""
        try:
            response = self.output_queue.get_nowait()
            self.logger.debug("Received new message: '%s'", response.text)

            # AIDEV-NOTE: Set state to waiting, then speaking
            self._set_state(SatelliteState.WAITING)

            # Get TTS audio synchronously
            tts_start_time = time.time()
            audio_bytes = await speech_recognition_tools.send_text_to_tts_api(response.text, self.config)
            tts_processing_time = time.time() - tts_start_time

            if audio_bytes:
                self._set_state(SatelliteState.SPEAKING)
                playback_start_time = time.time()
                # Calculate total processing time from stop sound to TTS playback start
                if self._stop_sound_time > 0:
                    total_processing_time = playback_start_time - self._stop_sound_time
                    self.logger.debug("Total processing time (STT + LLM + TTS): %.3f seconds", total_processing_time)

                self.logger.debug("Starting TTS playback at: %.3f", playback_start_time)
                # Play audio directly
                self.stream_output.write(audio_bytes)
                playback_duration = time.time() - playback_start_time
                self.logger.debug(
                    "Finished playing TTS audio (TTS processing: %.3f seconds, playback duration: %.3f seconds)",
                    tts_processing_time,
                    playback_duration,
                )

            # Return to listening state
            self._set_state(SatelliteState.LISTENING)

        except asyncio.QueueEmpty:
            pass  # No messages to process

    def _reset_audio_buffer(self) -> None:
        """Reset the audio buffer for new recording."""
        self._buffer_position = 0

    def _append_to_audio_buffer(self, audio_chunk: np.ndarray) -> bool:
        """Append audio chunk to buffer. Returns False if buffer is full."""
        chunk_size = len(audio_chunk)
        if self._buffer_position + chunk_size > len(self._audio_buffer):
            return False  # Buffer full

        self._audio_buffer[self._buffer_position : self._buffer_position + chunk_size] = audio_chunk
        self._buffer_position += chunk_size
        return True

    def _get_recorded_audio(self) -> np.ndarray:
        """Get the recorded audio."""
        if self._buffer_position == 0:
            return np.array([], dtype=np.float32)
        return self._audio_buffer[: self._buffer_position].copy()

    async def _record_command(self) -> None:
        """Record voice command after wake word detection."""
        self._set_state(SatelliteState.RECORDING)
        self._reset_audio_buffer()
        recording_start_time = time.time()

        try:
            silence_packages = 0
            max_frames = self.config.max_command_input_seconds * self.config.samplerate
            max_silent_packages = self.config.samplerate / self.config.chunk_size * self.config.max_length_speech_pause
            has_audio_data = False
            active_listening = True

            while active_listening:
                # Read directly from audio stream
                audio_bytes = self.stream_input.read(self.config.chunk_size, exception_on_overflow=False)

                if self.vad_model(audio_bytes) > self.config.vad_threshold:
                    # Convert audio to float32
                    raw_audio_data = np.frombuffer(audio_bytes, dtype=np.int16)
                    audio_chunk = raw_audio_data.astype(np.float32) / 32768.0

                    silence_packages = 0
                    if not self._append_to_audio_buffer(audio_chunk):
                        active_listening = False
                        self.logger.warning("Audio buffer full, stopping recording")
                    else:
                        has_audio_data = True
                    self.logger.debug("Received voice...")
                else:
                    if has_audio_data:
                        raw_audio_data = np.frombuffer(audio_bytes, dtype=np.int16)
                        audio_chunk = raw_audio_data.astype(np.float32) / 32768.0

                        if not self._append_to_audio_buffer(audio_chunk):
                            active_listening = False
                            self.logger.warning("Audio buffer full, stopping recording")
                        else:
                            silence_packages += 1
                    self.logger.debug("No voice...")

                if has_audio_data and (self._buffer_position > max_frames or silence_packages >= max_silent_packages):
                    active_listening = False
                    recording_duration = time.time() - recording_start_time
                    self.logger.debug("Stopping listening, playing stop sound...")
                    self.stream_output.write(self.stop_listening_sound)
                    self.logger.debug("Recording duration: %.3f seconds", recording_duration)

                    # Process STT
                    audio_frames = self._get_recorded_audio()
                    if len(audio_frames) > 0:
                        # Store stop sound time for total processing calculation
                        self._stop_sound_time = time.time()
                        await self._process_stt(audio_frames)

                await asyncio.sleep(0.001)

        finally:
            self._set_state(SatelliteState.LISTENING)

    async def _process_stt(self, audio_frames: np.ndarray) -> None:
        """Process speech-to-text and publish to MQTT."""
        self._set_state(SatelliteState.WAITING)
        stt_start_time = time.time()
        self.logger.debug("Processing STT request")

        try:
            response = await speech_recognition_tools.send_audio_to_stt_api(audio_frames, config_obj=self.config)
            stt_processing_time = time.time() - stt_start_time

            if response is not None:
                self.logger.debug(
                    "STT result: '%s' (processing time: %.3f seconds)", response.text, stt_processing_time
                )

                message = messages.ClientRequest(
                    id=uuid.uuid4(),
                    text=response.text,
                    room=self.config.room,
                    output_topic=self.config.output_topic,
                ).model_dump_json()

                # AIDEV-NOTE: Schedule MQTT publish on the MQTT thread's event loop
                mqtt_publish_start = time.time()
                if self._mqtt_loop and not self._mqtt_loop.is_closed():
                    future = asyncio.run_coroutine_threadsafe(
                        self.mqtt_client.publish(self.config.input_topic, message, qos=1), self._mqtt_loop
                    )
                    # Wait for the publish to complete
                    future.result(timeout=5)
                    mqtt_publish_time = time.time() - mqtt_publish_start
                    self.logger.debug("Published result text to MQTT (publish time: %.3f seconds).", mqtt_publish_time)
                else:
                    self.logger.error("MQTT loop not available, cannot publish STT result")
            else:
                self.logger.warning("STT processing failed for recorded audio")
        except Exception as e:
            self.logger.error("STT processing failed: %s", e)
        finally:
            self._set_state(SatelliteState.LISTENING)

    def _start_mqtt_loop(self) -> None:
        """Start the MQTT client in a separate thread."""

        def mqtt_runner():
            self._mqtt_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._mqtt_loop)
            try:
                self._mqtt_loop.run_until_complete(self.mqtt_client.start())
            except Exception as e:
                self.logger.error("MQTT client error: %s", e)
            finally:
                self._mqtt_loop.close()

        self._mqtt_thread = threading.Thread(target=mqtt_runner, daemon=True)
        self._mqtt_thread.start()
        time.sleep(1)  # Give MQTT client time to start

    async def _main_loop(self) -> None:
        """Main processing loop - handles wake word detection and state transitions."""
        try:
            while True:
                current_state = self._get_state()

                # Only process audio for wake word in LISTENING state
                if current_state == SatelliteState.LISTENING:
                    # Read audio chunk optimized for wake word detection
                    audio_data = self.stream_input.read(self.config.chunk_size_ow, exception_on_overflow=False)
                    audio_np = np.frombuffer(audio_data, dtype=np.int16)

                    # Wake word detection
                    prediction = self.wakeword_model.predict(
                        audio_np,
                        debounce_time=2.0,
                        threshold={self.config.name_wakeword_model: self.config.wakework_detection_threshold},
                    )
                    wakeword_probability = prediction[self.config.name_wakeword_model]

                    self.logger.debug(
                        "Wakeword probability: %s, Threshold: %s",
                        wakeword_probability,
                        self.config.wakework_detection_threshold,
                    )

                    if wakeword_probability >= self.config.wakework_detection_threshold:
                        self.logger.debug("Wakeword detected, playing start listening sound.")
                        self.stream_output.write(self.start_listening_sound)
                        await self._record_command()

                # Always check for TTS messages (low latency MQTT processing)
                await self._process_output_queue()

                # Small yield to prevent busy-waiting
                await asyncio.sleep(0.001)

        except KeyboardInterrupt:
            self.logger.info("Received interrupt signal")
            raise

    def start(self) -> None:
        """Start the satellite with simple state machine."""
        # Start MQTT client in separate thread
        self._start_mqtt_loop()

        # Run main processing loop
        try:
            asyncio.run(self._main_loop())
        except KeyboardInterrupt:
            self.logger.info("Received interrupt signal")
            raise

    def cleanup(self) -> None:
        """Clean up resources including MQTT client and audio streams."""
        if self._mqtt_thread and self._mqtt_thread.is_alive():
            self._mqtt_thread.join(timeout=2.0)

        self.stream_input.stop_stream()
        self.stream_output.stop_stream()
        self.stream_input.close()
        self.stream_output.close()
        self.p.terminate()



"""Basic tests for silero_vad module."""

from unittest.mock import MagicMock, patch

import pytest

from private_assistant_comms_satellite.silero_vad import SileroVad


class TestSileroVad:
    """Basic tests for SileroVad class."""

    @patch("private_assistant_comms_satellite.silero_vad.SileroVoiceActivityDetector")
    def test_silero_vad_initialization(self, mock_detector_class):
        """Test SileroVad initializes correctly."""
        mock_detector = MagicMock()
        mock_detector_class.return_value = mock_detector

        vad = SileroVad(threshold=0.5, trigger_level=2)

        assert vad.threshold == 0.5  # noqa: PLR2004
        assert vad.trigger_level == 2  # noqa: PLR2004
        assert vad._activation == 0
        assert vad.detector == mock_detector

    @patch("private_assistant_comms_satellite.silero_vad.SileroVoiceActivityDetector")
    def test_call_with_none_resets(self, mock_detector_class):
        """Test calling with None resets the detector."""
        mock_detector = MagicMock()
        mock_detector_class.return_value = mock_detector

        vad = SileroVad(threshold=0.5, trigger_level=2)
        vad._activation = 5  # Set some activation

        result = vad(None)

        assert result is False
        assert vad._activation == 0
        mock_detector.reset.assert_called_once()

    @patch("private_assistant_comms_satellite.silero_vad.SileroVoiceActivityDetector")
    def test_call_with_insufficient_audio_data(self, mock_detector_class):
        """Test calling with insufficient audio data raises ValueError."""
        mock_detector = MagicMock()
        mock_detector.chunk_bytes.return_value = 1024  # Require 1024 bytes
        mock_detector_class.return_value = mock_detector

        vad = SileroVad(threshold=0.5, trigger_level=2)

        # Provide less than required bytes
        with pytest.raises(ValueError, match="Audio bytes must be at least 1024 bytes"):
            vad(b"short_audio")  # Only 11 bytes

    @patch("private_assistant_comms_satellite.silero_vad.SileroVoiceActivityDetector")
    def test_call_speech_detection_below_threshold(self, mock_detector_class):
        """Test speech detection below threshold."""
        mock_detector = MagicMock()
        mock_detector.chunk_bytes.return_value = 10  # Small chunk for test
        mock_detector.return_value = 0.3  # Below threshold of 0.5
        mock_detector_class.return_value = mock_detector

        vad = SileroVad(threshold=0.5, trigger_level=2)

        # Provide sufficient audio data
        audio_data = b"a" * 20  # 20 bytes, enough for 2 chunks of 10
        result = vad(audio_data)

        assert result is False
        assert vad._activation == 0  # Should decrease/stay at 0

    @patch("private_assistant_comms_satellite.silero_vad.SileroVoiceActivityDetector")
    def test_call_speech_detection_above_threshold_not_triggered(self, mock_detector_class):
        """Test speech detection above threshold but not enough to trigger."""
        mock_detector = MagicMock()
        mock_detector.chunk_bytes.return_value = 10
        mock_detector.return_value = 0.7  # Above threshold of 0.5
        mock_detector_class.return_value = mock_detector

        vad = SileroVad(threshold=0.5, trigger_level=3)

        audio_data = b"a" * 20
        result = vad(audio_data)

        assert result is False
        assert vad._activation == 1  # Should increment but not trigger yet

    @patch("private_assistant_comms_satellite.silero_vad.SileroVoiceActivityDetector")
    def test_call_speech_detection_triggers(self, mock_detector_class):
        """Test speech detection that triggers activation."""
        mock_detector = MagicMock()
        mock_detector.chunk_bytes.return_value = 10
        mock_detector.return_value = 0.8  # Above threshold
        mock_detector_class.return_value = mock_detector

        vad = SileroVad(threshold=0.5, trigger_level=2)
        vad._activation = 1  # Already at activation level 1

        audio_data = b"a" * 20
        result = vad(audio_data)

        assert result is True
        assert vad._activation == 0  # Should reset after trigger

    @patch("private_assistant_comms_satellite.silero_vad.SileroVoiceActivityDetector")
    def test_call_multiple_chunks_max_probability(self, mock_detector_class):
        """Test that multiple chunks use maximum probability."""
        mock_detector = MagicMock()
        mock_detector.chunk_bytes.return_value = 5
        # Return different probabilities for different chunks
        mock_detector.side_effect = [0.3, 0.8, 0.4]  # Max is 0.8
        mock_detector_class.return_value = mock_detector

        vad = SileroVad(threshold=0.5, trigger_level=1)

        audio_data = b"a" * 15  # 15 bytes = 3 chunks of 5
        result = vad(audio_data)

        assert result is True  # Should trigger because max(0.3, 0.8, 0.4) = 0.8 > 0.5
        assert mock_detector.call_count == 3  # noqa: PLR2004 - Should process 3 chunks



"""Tests for speech_recognition_tools module."""

import numpy as np
import pytest
from pydantic import ValidationError

from private_assistant_comms_satellite.utils import config, speech_recognition_tools


@pytest.fixture
def sample_config():
    """Create a sample config for testing."""
    return config.Config(
        speech_transcription_api="http://test-stt.example.com/transcribe",
        speech_transcription_api_token="test-stt-token",
        speech_synthesis_api="http://test-tts.example.com/synthesize",
        speech_synthesis_api_token="test-tts-token",
    )


@pytest.fixture
def sample_audio_data():
    """Create sample audio data for testing."""
    # Create a simple sine wave as test audio
    sample_rate = 16000
    duration = 1.0  # 1 second
    frequency = 440  # A4 note
    t = np.linspace(0, duration, int(sample_rate * duration), False)
    return np.sin(2 * np.pi * frequency * t).astype(np.float32)


class TestInt2Float:
    """Test the int2float conversion function."""

    def test_int2float_with_valid_data(self):
        """Test int2float with valid int16 data."""
        # Create test data
        int_data = np.array([1000, -1000, 0, 16383, -16384], dtype=np.int16)

        result = speech_recognition_tools.int2float(int_data)

        assert result.dtype == np.float32
        assert len(result) == len(int_data)
        # Check that values are normalized to [-1, 1] range
        assert np.all(np.abs(result) <= 1.0)

    def test_int2float_with_zeros(self):
        """Test int2float with all zeros."""
        int_data = np.array([0, 0, 0], dtype=np.int16)

        result = speech_recognition_tools.int2float(int_data)

        assert result.dtype == np.float32
        assert np.all(result == 0.0)

    def test_int2float_squeeze_behavior(self):
        """Test that int2float properly squeezes dimensions."""
        # Create 2D array that should be squeezed
        int_data = np.array([[1000], [2000], [3000]], dtype=np.int16)

        result = speech_recognition_tools.int2float(int_data)

        assert result.ndim == 1  # Should be squeezed to 1D
        assert len(result) == 3  # noqa: PLR2004


# NOTE: Async API tests removed for simplicity - focus on basic unit tests


class TestSTTResponse:
    """Test the STTResponse model."""

    def test_valid_stt_response(self):
        """Test creating STTResponse with valid data."""
        data = {"text": "hello world", "message": "success"}

        response = speech_recognition_tools.STTResponse.model_validate(data)

        assert response.text == "hello world"
        assert response.message == "success"

    def test_invalid_stt_response_missing_text(self):
        """Test STTResponse validation with missing text field."""
        data = {"message": "success"}  # Missing 'text' field

        with pytest.raises(ValidationError):
            speech_recognition_tools.STTResponse.model_validate(data)

    def test_invalid_stt_response_missing_message(self):
        """Test STTResponse validation with missing message field."""
        data = {"text": "hello world"}  # Missing 'message' field

        with pytest.raises(ValidationError):
            speech_recognition_tools.STTResponse.model_validate(data)

    def test_stt_response_extra_fields(self):
        """Test STTResponse with extra fields (should be ignored)."""
        data = {"text": "hello world", "message": "success", "extra_field": "ignored"}

        response = speech_recognition_tools.STTResponse.model_validate(data)

        assert response.text == "hello world"
        assert response.message == "success"
        # Extra field should not be accessible
        assert not hasattr(response, "extra_field")
